#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the SCUT_FBP5500 dataset."""

from tensorbay.opendataset.SCUT_FBP5500.loader import SCUT_FBP5500

__all__ = ["SCUT_FBP5500"]
