#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the Animals with attributes 2 dataset."""

from tensorbay.opendataset.AnimalsWithAttributes2.loader import AnimalsWithAttributes2

__all__ = ["AnimalsWithAttributes2"]
