#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of VOC2012Detection."""

from tensorbay.opendataset.VOC2012Detection.loader import VOC2012Detection

__all__ = ["VOC2012Detection"]
