#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the SegTrack2 dataset."""

from tensorbay.opendataset.SegTrack2.loader import SegTrack2

__all__ = ["SegTrack2"]
