#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the OxfordIIITPet dataset."""

from tensorbay.opendataset.OxfordIIITPet.loader import OxfordIIITPet

__all__ = ["OxfordIIITPet"]
