#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the WIDER FACE dataset."""

from tensorbay.opendataset.WIDER_FACE.loader import WIDER_FACE

__all__ = ["WIDER_FACE"]
