#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the LISA traffic light dataset."""

from tensorbay.opendataset.LISATrafficLight.loader import LISATrafficLight

__all__ = ["LISATrafficLight"]
