#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the CompCars dataset."""

from tensorbay.opendataset.CompCars.loader import CompCars

__all__ = ["CompCars"]
