#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the Hard Hat Workers dataset."""

from tensorbay.opendataset.HardHatWorkers.loader import HardHatWorkers

__all__ = ["HardHatWorkers"]
