#!/usr/bin/env python3
#
# Copytright 2020 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of UAVDT."""

from tensorbay.opendataset.UAVDT.loader import UAVDT

__all__ = ["UAVDT"]
