#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the The Car Connection Picture dataset."""

from tensorbay.opendataset.CarConnection.loader import CarConnection

__all__ = ["CarConnection"]
