#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the COCO2017 dataset."""

from tensorbay.opendataset.COCO2017.loader import COCO2017

__all__ = ["COCO2017"]
