#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the RP2K dataset."""

from tensorbay.opendataset.RP2K.loader import RP2K

__all__ = ["RP2K"]
