#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloaders of the BDD100K_MOT2020 dataset and the BDD100K_MOTS2020 dataset."""

from tensorbay.opendataset.BDD100K_MOT2020.loader import BDD100K_MOT2020, BDD100K_MOTS2020

__all__ = ["BDD100K_MOT2020", "BDD100K_MOTS2020"]
