#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the JHU-CROWD++ dataset."""

from tensorbay.opendataset.JHU_CROWD.loader import JHU_CROWD

__all__ = ["JHU_CROWD"]
