#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the Newsgroups20 dataset."""

from tensorbay.opendataset.Newsgroups20.loader import Newsgroups20

__all__ = ["Newsgroups20"]
