#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the DownsampledImagenet dataset."""

from tensorbay.opendataset.DownsampledImagenet.loader import DownsampledImagenet

__all__ = ["DownsampledImagenet"]
