#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the Halpe Full-Body Human Keypoints and HOI-Det dataset."""

from tensorbay.opendataset.HalpeFullBody.loader import HalpeFullBody

__all__ = ["HalpeFullBody"]
