#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the COVID_CT dataset."""

from tensorbay.opendataset.COVID_CT.loader import COVID_CT

__all__ = ["COVID_CT"]
