#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of CCPD."""

from tensorbay.opendataset.CCPD.loader import CCPD, CCPDGreen

__all__ = ["CCPD", "CCPDGreen"]
